#! python3
'''
   xy.py: Module for XyPy|XPyL (Python embedded in XPL)
   C.L.Distefano rev. 2015-10-04
   ----------------------------------------------------------
   Requires XyWrite 4 (DOS) + XyWWWeb.U2 + Python 3.x.
   See XYPY.TXT (in http://users.datarealm.com/xywwweb/XYPY.ZIP)
     for setup instructions and usage.

'''

import os

# Global variables
# ----------------
EDDIR = '' # Editor's directory (set by U2 frame MAKEXYPY)
CMFN = EDDIR + '/cml.sav'
INFN = EDDIR + '/xpyldata.in'   # Input from XyWrite
OUTFN = EDDIR + '/xpyldata.out' # Output from Python script
OKFN = EDDIR + '/xpyldata.ok'   # Output ready flag file
bLG = b'\xAE' # byte code for left guillemet
bRG = b'\xAF' # byte code for right guillemet

# 3-byte Function mnemonics
FUNCNAMES = [
'@0', '@1', '@2', '@3', '@4', '@5', '@6', '@7', '@8', '@9', '@A', '@B',
'@C', '@D', '@E', '@F', '@G', '@H', '@I', '@J', '@K', '@L', '@M', '@N',
'@O', '@P', '@Q', '@R', '@S', '@T', '@U', '@V', '@W', '@X', '@Y', '@Z',
'AD', 'AS', 'BF', 'BK', 'BS', 'CC', 'CD', 'CH', 'CI', 'CL', 'CM', 'CN',
'CP', 'CR', 'CS', 'CU', 'DC', 'DF', 'GH', 'DL', 'DP', 'DS', 'DW', 'EL',
'ER', 'EX', 'GT', 'HM', 'M0', 'M1', 'M2', 'M3', 'M4', 'M5', 'M6', 'M7',
'M8', 'MD', 'MU', 'MV', 'NC', 'NL', 'NK', 'NP', 'NR', 'NS', 'NT', 'NW',
'PC', 'PD', 'PL', 'PP', 'PR', 'PS', 'PT', 'PU', 'PW', 'R0', 'R1', 'R2',
'R3', 'R4', 'R5', 'R6', 'R7', 'R8', 'R9', 'RC', 'RD', 'RE', 'RL', 'RP',
'RS', 'RV', 'RW', 'SD', 'SH', 'SI', 'SK', 'SM', 'SN', 'SS', 'SU', 'SV',
'TF', 'TI', 'TN', 'TS', 'UD', 'WA', 'WC', 'WC', 'WN', 'WS', 'WX', 'WW',
'XC', 'XD', 'DT', 'S1', 'S2', 'S3', 'S4', 'S5', 'S6', 'S7', 'SP', 'BC',
'LB', 'LE', 'NF', 'PF', 'TP', 'BD', 'MS', 'NM', 'LD', 'LL', 'LR', 'LU',
'UP', 'FF', 'YD', 'DO', 'DX', 'MK', 'SO', 'OP', 'WZ', 'NX', 'SW', 'FD',
'FM', 'TL', 'TR', 'TE', 'ED', 'EE', 'HC', 'EC', 'MC', '#1', '#2', '#3',
'#4', '#5', '#6', '#7', '#8', '#9', '$1', '$2', '$3', '$4', '$5', '$6',
'$7', '$8', '$9', 'DR', 'EN', 'C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6',
'C7', 'C8', 'C9', 'EF', 'IB', 'NO', 'NI', 'CO', '$0', 'LS', 'XP', 'WG',
'XM', '&0', '&1', '&2', '&3', '&4', '&5', '&6', '&7', '&8', '&9', '&A',
'&B', '&C', '&D', '&E', '&F', '&G', '&H', '&I', '&J', '&K', '&L', '&M',
'&N', '&O', '&P', '&Q', '&R', '&S', '&T', '&U', '&V', '&W', '&X', '&Y',
'&Z', 'HL', '$A', '$B', '$C', '$D', '$E', '$F', '$G', '$H', '$I', '$J',
'$K', '$L', '$M', '$N', '$O', '$P', '$Q', '$R', '$S', '$T', '$U', '$V',
'$W', '$X', '$Y', '$Z', 'XX', 'H@', 'VH', 'MW', 'QH', 'DK', 'SR', 'SC',
'TG', 'H1', 'JH', 'DZ', 'DD', 'DM', 'LT', 'RK', 'NN', 'MT', 'ET', 'ZT',
'T1', 'TT', '<<', '>>', 'IT', 'SL', 'SF', 'FL', 'FR', 'FC', 'SY', 'ME',
'AC', 'FS', 'TW', 'MI', 'RO', 'NB', 'Q1', 'Q2', 'Q3', 'Q4', 'Q5', 'Q6',
'Q7', 'Q8', 'TO', 'IR', 'AR', 'AX', 'DB', 'DE', 'HF', 'SA', 'OV', 'TC',
'TB', 'JM', 'SG', 'XH', 'FT', 'BX', 'MN', 'CB', 'M9', 'MZ', 'ZZ', 'RX',
'ST', 'KF', 'JC', 'AK', 'TM', 'NU', 'B4', 'QP', 'HG', 'US', 'XE', 'ES',
'RB', 'S-', 'S+', '**', 'BN', 'RU', 'CF', 'UI', 'XS', 'EA', 'BT', 'KD',
'DN', 'HI', 'WH', 'XN', 'FX', 'UN', 'MX', 'AZ', 'BR', 'HK', '#X', '#X',
'BM', 'JR', 'XO', 'XW', 'TX', 'LF', 'LO', 'BL', 'XT', 'WT', 'IC', 'CT',
'VB', '-D', 'WD', 'RM', 'LM', 'aL', 'aR', 'aB', 'aE', 'MP', 'mN', 'QL',
'QR', 'MF']

# Function definitions
# --------------------
def outok(flag = False, okfile = OKFN):
  ''' Output ready flag file (0=write|1=delete, flag_filename) '''
  if not flag:
    with open(okfile, 'wt') as okfo:
      okfo.write('')
  else:
    try:
      os.remove(okfile)
    except FileNotFoundError:
      pass

def xyread(file_in = INFN, mode = 'rb'):
  ''' Read XyWrite input file (minus EOF) '''
  try:
    with open(file_in, mode) as f:
      fb = f.read()
      if fb[-1] == 26: fb = fb[:-1] # fails with 0-byte file
      return fb
  except: return None

def xywrite(bytes_in, file_out = OUTFN, mode = 'wb'):
  ''' Write file for output to XyWrite '''
  try:
    with open(file_out, mode) as f:
      f.write(bytes_in)
      return 1
  except: return None

def GetCMline():
  ''' Fetch CMline file in binary mode '''
  with open(CMFN, 'rb') as f:
    cmline = f.read()
    if cmline[-1] == 26: # fails with 0-byte file
      cmline = cmline[:-1]
    return cmline

def xycm(*cm_in):
  ''' Returns bytes for XyWrite embedded command(s) '''
  ba = bytearray()
  for cm in cm_in:
    if isinstance(cm, str):
      cm = cm.encode()
    try:
      ba.extend(b'\xae' + cm + b'\xaf')
    except ValueError:
      print('Bad:', cm, '- must be Ascii text string or bytes')
      return None
    except Exception as e:
      print(e)
      return None
  return ba
    
def fBX(*cm_in):
  ''' Returns bytes for BX [command]Q2 '''
  ba = bytearray()
  for cm in cm_in:
    if isinstance(cm, str):
      cm = cm.encode()
    try:
      ba.extend(b'\xff\x82\xab' + cm + b'\xff\x82\x7f')
    except ValueError:
      print('Bad:', cm, '- must be Ascii text string or bytes')
      return None
    except Exception as e:
      print(e)
      return None
  return ba
   
def xyfunc(*func_mnemonic):
  ''' Return bytes for XyWrite 3-byte function mnemonic(s) '''
  ba = bytearray()
  for mn in func_mnemonic:
    try:
      if mn not in FUNCNAMES: mn = mn.upper()
      n = FUNCNAMES.index(mn)
      #        byte1 byte2            byte3
      for b in 255,  128 + n // 128,  (1 + n * 2) % 256:
        ba.append(b)
    except Exception:
      print(mn, 'is not a valid 3-byte func mnemonic')
      return None
  return ba
  
def xywild(*char_in):
  ''' Returns bytes for standard reverse-video wildcard character(s)
       by normal Ascii char or Ascii-number (0 - 255) '''
  ba = bytearray()
  for n in char_in:
    if isinstance(n, str):
      if len(n) == 1: n = ord(n)
    if isinstance(n, int):
      if 0 <= n < 256:
        n = (n + 128) % 256
        b2 = 192
        if n < 128: b2 = 193
        for v in 255, b2, n:
          ba.append(v)
    else:
      raise ValueError('1-byte Ascii char or int 0-255 only')
      return None
  return ba
  
def xy3bchar(*char_in):
  ''' Returns bytes for XyWrite 3-byte printable char(s) (0 -255)
      args = 1-byte Ascii char or integer 0 - 255 
  '''
  ba = bytearray()
  for n in char_in:
    if isinstance(n, str):
      if len(n) == 1: n = ord(n)
    if isinstance(n, int):
      #   next line converts, e.g., 0xa to '0A'
      h = '{0:0>2s}'.format(hex(n).split('x')[1].upper())
      #         byte1 byte2       byte3
      for bt in 255,  ord(h[0]),  ord(h[1]):
        ba.append(bt)
    else:
      raise ValueError('1-byte Ascii 0 - 255 char(s) only')
      return None    
  return ba

def xychar(*char_num):
  ''' Return bytes for XyWrite 3-byte Speedo char(s) 256 - 1023 '''
  ba = bytearray()
  for n in char_num:
    if isinstance(n, int):
      if 256 <= n < 1024:
        #        byte1 byte2             byte3
        for b in 254,  (n - 256) % 256,  (n - 256) // 256:
          ba.append(b)
      else:
        raise ValueError(errmsg)
        return None        
    else:
      raise ValueError(errmsg)
      return None
  return ba

def GetStackall():
  ''' Read command archive file STACKALL.SAV '''
  stackall = None
  with open(EDDIR + '/stackall.sav', 'rb') as f:
    stackall = f.read()
  try:
    if stackall[-1] == 26:
      stackall = stackall[:-1]
  except: pass
  return stackall

def PutStackall(bytes_in):
  ''' Write command archive file STACKALL.SAV '''
  if bytes_in:
    with open(EDDIR + '/stackall.sav', 'wb') as f:
      try:
        f.write(bytes_in)
        return 1
      except: pass
  else: pass

def bconv(num, base_in = 16, base_out = 10):
  ''' Base-to-base number conversion '''
  prefix = ''
  try:
    num = int(str(num), base_in)
    if base_out == 10: return num
    if str(num).startswith('-'):
      num = str(num)[1:]
      prefix = '-'
    L = '0123456789abcdefghijklmnopqrstuvwxyz'
    num = int(num)
    digits_out = '' # convert base10 to base_out value
    while num >= base_out:
      i = divmod(num, base_out)
      digits_out += L[:base_out][i[1]]
      num = i[0]
    digits_out += L[:base_out][num]
    return prefix + digits_out[::-1]
  except Exception as e:
    print(e)
    return None

if __name__ == '__main__':
  print(__doc__)

# end xy.py